#!/usr/bin/env node

import Analyze from '../commands/analyze'
import Serve from '../commands/serve'
import Build from '../commands/build'
import Setup from '../commands/setup'
import Page from '../commands/page'
import { relative } from 'path'
import neodoc from 'neodoc'

const cli = `
  🐙  Elmo – Dynamic websites on static infrastructure.

  Usage:

    elmo [options]
    elmo [options] setup [<dir>]
    elmo [options] page <page>
    elmo [options] build [<out>]
    elmo [options] serve [<port>]
    elmo [options] analyze

  Commands:

    setup:     scaffold a new project
    page:      scaffold a new page
    build:     build the project
    serve:     start the development server
    analyze:   print details about the build

  Global Options:

    --typescript   Setup a typescript project
    --dir <dir>    Working directory
    --dev          Development build
    --h, --help    Print this screen
    --version      Get the version
`

main().catch(err => {
  console.error(err.stack || err.message)
  process.exit(1)
})

// TODO: get more specific
type Args = {
  [key: string]: string
}

async function main() {
  const args: Args = neodoc.run(cli, { smartOptions: true, laxPlacement: true })

  if (args.setup) {
    return Setup({
      root: args['<dir>'] || args['--dir'],
      typescript: !!args['--typescript']
    })
  }

  if (args.page) {
    await Page({
      root: args['--dir'],
      pageName: args['<page>'],
      typescript: !!args['--typescript']
    })
    console.error('🐙  wrote: ./%s', args['<page>'])
    return
  }

  if (args.build) {
    return Build({
      root: args['--dir'],
      environment: args['--dev'],
      outpath: args['--out']
    })
  }

  if (args.serve) {
    await Serve({
      root: args['--dir'],
      environment: args['--dev'],
      port: parseInt(args['<port>'], 10)
    })

    const rel = relative(process.cwd(), args['--dir'])
    if (rel) {
      console.error(
        `🐙  All set! Run \`cd ${rel} && elmo serve\` to get started.`
      )
    } else {
      console.error(`🐙  All set! Run \`elmo serve\` to get started.`)
    }

    return
  }

  if (args.analyze) {
    return await Analyze({
      root: args['--dir'],
      environment: args['--dev']
    })
  }

  console.error(cli)
  return
}
