/*
 * Module dependencies
 */

import Config from '../server/config/webpack'
import * as move from 'move-concurrently'
import { join, resolve } from 'path'
import * as Webpack from 'webpack'
import { directory } from 'tempy'
import * as copy from 'cpy'
import * as del from 'del'

export type BuildOptions = {
  root?: string
  outpath?: string
  environment?: string
}

export default async function Build(
  params: BuildOptions
): Promise<Webpack.Stats[]> {
  const env = params.environment || 'production'
  const root = params.root || process.cwd()
  const tmpdir = directory()
  const outpath = resolve(root, params.outpath || 'build')

  // start loading the server-side config
  const s = Config({
    environment: env,
    target: 'server',
    action: 'build',
    outpath: tmpdir,
    routes: [],
    root
  })

  // start loading the client-side config
  const b = Config({
    environment: env,
    target: 'browser',
    action: 'build',
    outpath: tmpdir,
    routes: [],
    root
  })

  // build the configurations in parallel
  const stats = await Promise.all(
    [s, b].map(async config => {
      const c: Webpack.Configuration = await config
      const webpack = Webpack(c)
      return new Promise<Webpack.Stats>((resolve, reject) => {
        webpack.run((err: Error, stats: Webpack.Stats) => {
          if (err) return reject(err)

          if (stats.hasErrors()) {
            return reject(new Error(stats.toJson().errors.join('\n\n')))
          }

          return resolve(stats)
        })
      })
    })
  )

  // copy the _redirects file over if there is any
  await copy(join(root, '_redirects'), tmpdir)

  // delete the existing outpath
  await del(outpath)

  // move the files over
  await move(tmpdir, outpath)

  return stats
}
