//

/*
 * Module dependencies
 */

import * as WebpackDevMiddleware from 'webpack-dev-middleware'
import * as WebpackHotMiddleware from 'webpack-hot-middleware'
import Config from '../server/config/webpack'
import Routes from '../server/routes'
import * as Deferred from 'deferral'
import * as Webpack from 'webpack'
import * as express from 'express'
import { resolve } from 'path'

const regexps: { [s: string]: RegExp } = {}
const app = express()

export type ServeOptions = {
  root?: string
  port?: number
  environment?: string
}

export default async function serve(params: ServeOptions) {
  params.root = params.root || process.cwd()
  const outpath = resolve(params.root, 'build')
  const port: number = params.port || 3130
  const routes = await Routes(params.root)

  // start loading the server-side config
  const s = Config({
    environment: params.environment || 'development',
    outpath: outpath,
    root: params.root,
    target: 'server',
    action: 'serve',
    routes: routes
  })

  // start loading the client-side config
  const b = Config({
    environment: params.environment || 'development',
    outpath: outpath,
    root: params.root,
    target: 'browser',
    action: 'serve',
    routes: routes
  })

  // wait for both to finish
  const [server, browser] = await Promise.all([s, b])
  const firstBuild = new Deferred()

  // create the server-side watcher
  // this will also initiate a build
  // TODO: watcher will race against the browser build.
  // see if I can synchronize them, by first figuring out
  // when webpack starts recompiling and hooking into the
  // hot reload
  const watcher = Webpack(server).watch(
    {
      ignored: outpath
    },
    (err, stats) => {
      if (err) {
        console.error(err)
        return firstBuild.reject(err)
      }
      firstBuild.resolve(stats)
    }
  )

  // wait for the first build to complete
  // and don't go any further if there's
  // a build error
  await firstBuild

  // create the browser build
  const webpack = Webpack(browser)

  app.use(
    WebpackDevMiddleware(webpack, {
      publicPath: '/',
      logLevel: 'silent',
    })
  )

  app.use(
    WebpackHotMiddleware(webpack, {
      log: false
    })
  )

  // support custom routing
  if (routes.length) {
    // rehydrate regexps
    app.use(function (req, _res, next) {
      if (req.method !== 'GET') return next()
      for (let i = 0, len = routes.length; i < len; i++) {
        const route = routes[i]
        const regexp = (regexps[route.regexp] =
          regexps[route.regexp] || new RegExp(`${route.regexp}`))
        if (!regexp.test(req.path)) continue
        req.url = route.path.replace(/\/+$/, '') + '/'
        return next()
      }
      return next()
    })
  }

  app.use(express.static(resolve(params.root, outpath || 'build')))

  // start listening
  const listening = new Deferred()
  const http = app.listen(port)
  http.once('error', err => listening.resolve(err))
  const err: Error = await listening

  // cleanup, not sure if we need to wait
  // for it to finish or not
  const closing = new Deferred()
  watcher.close(() => closing.resolve())
  await closing

  // throw the message again after cleanup
  throw err
}
