import { join, basename, dirname } from 'path'
import { exists, writeFile } from 'mz/fs'
import * as readPkg from 'read-pkg-up'
import * as mkdir from 'make-dir'
import * as copy from 'cpy'
import { set } from 'shvl'

export type SetupOptions = {
  root?: string
  typescript: boolean
}

export default async function setup(params: SetupOptions) {
  const { path: elmoPkgPath } = await readPkg({ cwd: __dirname })
  const elmoDir = dirname(elmoPkgPath)
  const scaffoldDir = join(elmoDir, "scaffold")

  const root = params.root || process.cwd()
  const pages = join(root, 'pages')
  const statics = join(root, 'static')
  const types = join(root, 'types')

  // TODO: parallel
  await mkdir(pages)
  await mkdir(statics)
  await mkdir(types)
  await copy(`${scaffoldDir}/_redirects`, root, { overwrite: false })
  await copy(`${scaffoldDir}/static/**/*`, statics, { overwrite: false })

  if (params.typescript) {
    await copy(`${scaffoldDir}/pages/*.tsx`, pages, { overwrite: false })
    await copy(`${scaffoldDir}/pages/index/*.tsx`, join(pages, 'index'), { overwrite: false })
    await copy(`${scaffoldDir}/tsconfig.json`, root, { overwrite: false })
  } else {
    await copy(`${scaffoldDir}/pages/*.js`, pages, { overwrite: false })
    await copy(`${scaffoldDir}/pages/index/*.js`, join(pages, 'index'), { overwrite: false })
  }

  // create a package.json with some sensible
  // defaults based on the versions in elmo.
  const {
    version: ver,
    dependencies: { preact: preactVer }
   } = require(elmoPkgPath)

  // create the package.json if it doesn't yet exist
  const rootPkgPath = join(root, 'package.json')
  if (!await exists(rootPkgPath)) {
    await writeFile(
      rootPkgPath,
      `{
  "name": "${basename(root)}",
  "version": "1.0.0",
  "private": true,
  "dependencies": {
    "preact": "${preactVer}",
    "elmo": "${ver}"
  }
}`
    )
  } else {
    // let's ensure that we've got a matching version of preact set
    const pkgPath = require(rootPkgPath)
    set(pkgPath, "dependencies.preact", preactVer)
    set(pkgPath, "dependencies.elmo", ver)
    await writeFile(rootPkgPath, JSON.stringify(pkgPath, null, 2))
  }

  // setup the index page
  // await Page({ root, typescript: params.typescript, pageName: 'index' })
}
