import { h, Component, ComponentProps } from 'preact' // eslint-disable-line

export interface DocumentProps extends ComponentProps {
  filepath: string
  head: any
  style: any
  page: any
  props: any
}

export default class Document extends Component<DocumentProps, null> {
  static Head: typeof Head
  static Page: typeof Page
  static Scripts: typeof Scripts

  props: DocumentProps

  getChildContext() {
    return { _documentProps: this.props }
  }

  render() {
    return (
      <html>
        <Head>
          <meta name="viewport" content="width=device-width, initial-scale=1" />
          <link
            rel="shortcut icon"
            type="image/x-icon"
            href="/static/favicon.ico"
          />
          <meta name="fragment" content="!" />
        </Head>
        <body>
          <Page />
          <Scripts />
        </body>
      </html>
    )
  }
}

export class Head extends Component<any, never> {
  context: {
    _documentProps?: DocumentProps
  }

  render({ children }: ComponentProps) {
    if (!children) return null
    // prepend children if there's <Head />'s in the tree
    if (!this.context._documentProps) return <head>{children}</head>
    const { head, style } = this.context._documentProps
    if (head) {
      children = (head || []).concat(children).filter(unique()).reverse()
    }
    if (style) children = (children || []).concat(style)
    return <head>{children || []}</head>
  }
}

export class Page extends Component<Object, Object> {
  context: {
    _documentProps: DocumentProps
  }

  render() {
    const { page } = this.context._documentProps
    return <div id="elmo.page" dangerouslySetInnerHTML={{ __html: page }} />
  }
}

export class Scripts extends Component<Object, any> {
  context: {
    _documentProps: DocumentProps
  }

  render() {
    const { filepath } = this.context._documentProps
    return (
      <div id="elmo.scripts">
        <script
          type="text/javascript"
          dangerouslySetInnerHTML={{ __html: 'window.prerenderReady=false;' }}
        />
        <script type="text/javascript" src="/common.js" />
        <script type="text/javascript" src={`/${filepath}`} />
      </div>
    )
  }
}

const METATYPES = ['name', 'httpEquiv', 'charSet', 'itemProp']

// returns a function for filtering head child elements
// which shouldn't be duplicated, like <title/>.
// TODO: typescript
function unique() {
  const tags: any = []
  const metaTypes: any = []
  const metaCategories: any = {}
  return (h: any) => {
    switch (h.nodeName) {
      case 'title':
      case 'base':
        if (~tags.indexOf(h.nodeName)) return false
        tags.push(h.nodeName)
        break
      case 'meta':
        for (let i = 0, len = METATYPES.length; i < len; i++) {
          const metatype = METATYPES[i]
          if (!h.attributes.hasOwnProperty(metatype)) continue
          if (metatype === 'charSet') {
            if (~metaTypes.indexOf(metatype)) return false
            metaTypes.push(metatype)
          } else {
            const category = h.attributes[metatype]
            const categories = metaCategories[metatype] || []
            if (~categories.indexOf(category)) return false
            categories.push(category)
            metaCategories[metatype] = categories
          }
        }
        break
    }
    return true
  }
}

// Add in the static references
Document.Head = Head
Document.Page = Page
Document.Scripts = Scripts