import { Component, cloneElement } from 'preact'

const DOMAttributeNames: {
  [attr: string]: string
} = {
    acceptCharset: 'accept-charset',
    className: 'class',
    htmlFor: 'for',
    httpEquiv: 'http-equiv'
  }

const browser = typeof window !== 'undefined'
let mounted: Component<any, any>[] = []

function reducer(components: preact.ComponentProps[]) {
  return components
    .map(c => c.children)
    .filter((c: preact.PreactChildren) => !!c)
    .reduce((a: any, b: any) => a.concat(b), [])
    .reverse()
    .filter(unique())
    .reverse()
    .map((v: preact.PreactChild) => {
      if (typeof v === 'string') return v
      if (!('nodeName' in v)) return v
      const vnode = v as preact.VNode<any>
      const className =
        (vnode.attributes && vnode.attributes.className
          ? vnode.attributes.className + ' '
          : '') + 'preact-head'
      return cloneElement(vnode, { className })
    })
}

function updateClient(head: preact.VNode<any>[]) {
  const tags: {
    [nodeName: string]: preact.VNode<any>[]
  } = {}

  head.forEach(h => {
    // TODO: perhaps support custom components like <Meta />
    if (typeof h.nodeName !== 'string') return
    const components = tags[h.nodeName] || []
    components.push(h)
    tags[h.nodeName] = components
  })

  updateTitle(tags.title ? tags.title[0] : null)

  const types = ['meta', 'base', 'link', 'style', 'script']
  types.forEach(type => {
    updateElements(type, tags[type] || [])
  })
}

function updateElements(type: string, components: preact.VNode<any>[]) {
  const headEl = document.getElementsByTagName('head')[0]
  const oldTags = Array.prototype.slice.call(
    headEl.querySelectorAll(type + '.preact-head')
  )
  const newTags = components.map(VNodeToDOM).filter(newTag => {
    for (let i = 0, len = oldTags.length; i < len; i++) {
      const oldTag = oldTags[i]
      if (oldTag.isEqualNode(newTag)) {
        oldTags.splice(i, 1)
        return false
      }
    }
    return true
  })

  oldTags.forEach(
    (t: HTMLElement) => t.parentNode && t.parentNode.removeChild(t)
  )

  newTags.forEach((t: HTMLElement) => headEl.appendChild(t))
}

function VNodeToDOM(component: preact.VNode<any>) {
  const el = document.createElement(component.nodeName as string)
  const attrs = component.attributes || {}
  const children = component.children

  for (const p in attrs) {
    if (!attrs.hasOwnProperty(p)) continue
    if (p === 'dangerouslySetInnerHTML') continue

    const attr = DOMAttributeNames[p] || p.toLowerCase()
    el.setAttribute(attr, attrs[p])
  }

  if (attrs['dangerouslySetInnerHTML']) {
    el.innerHTML = attrs['dangerouslySetInnerHTML'].__html || ''
  } else if (children) {
    el.textContent = typeof children === 'string' ? children : children.join('')
  }

  return el
}

const METATYPES = ['name', 'httpEquiv', 'charSet', 'itemProp']

// returns a function for filtering head child elements
// which shouldn't be duplicated, like <title/>.
// TODO: typescript
function unique() {
  const tags: any = []
  const metaTypes: any = []
  const metaCategories: any = {}
  return (h: any) => {
    switch (h.nodeName) {
      case 'title':
      case 'base':
        if (~tags.indexOf(h.nodeName)) return false
        tags.push(h.nodeName)
        break
      case 'meta':
        for (let i = 0, len = METATYPES.length; i < len; i++) {
          const metatype = METATYPES[i]
          if (!h.attributes.hasOwnProperty(metatype)) continue
          if (metatype === 'charSet') {
            if (~metaTypes.indexOf(metatype)) return false
            metaTypes.push(metatype)
          } else {
            const category = h.attributes[metatype]
            const categories = metaCategories[metatype] || []
            if (~categories.indexOf(category)) return false
            categories.push(category)
            metaCategories[metatype] = categories
          }
        }
        break
    }
    return true
  }
}

function updateTitle(component: preact.VNode<any> | null) {
  let title
  if (component) {
    const { children } = component
    title = typeof children === 'string' ? children : children.join('')
  } else {
    title = ''
  }
  if (title !== document.title) {
    document.title = title
  }
}

function update() {
  if (!browser) return
  const state = reducer(mounted.map(mount => mount.props))
  updateClient(state)
}

export default class Head extends Component<any, any> {
  static rewind() {
    const children = reducer(mounted.map(mount => mount.props))
    mounted = []
    return children
  }

  componentDidUpdate() {
    update()
  }

  componentWillMount() {
    mounted.push(this)
    update()
  }

  componentWillUnmount() {
    const i = mounted.indexOf(this)
    if (~i) mounted.splice(i, 1)
    update()
  }

  render() {
    return null
  }
}
