//

/*
 * Module dependencies
 */

// import // development as babelDev,
// production as babelPro
// typescript as babelTS
// './babel'
import { Params } from '../../elmo'
import StaticSiteGeneratorPlugin from '../plugins/static'
import * as FriendlyErrors from 'friendly-errors-webpack-plugin'
import * as CopyWebpackPlugin from 'copy-webpack-plugin'
import * as readPkg from 'read-pkg-up'

// import { CheckerPlugin } from 'awesome-typescript-loader'
import {
  join,
  resolve,
  basename,
  dirname,
  // relative,
  extname
  // sep as PATH_SEP
} from 'path'
import { exists } from 'mz/fs'
import Route from '../../lib/route'
import * as webpack from 'webpack'
import * as globby from 'globby'
import Routes from '../routes'

/**
 * Environment Variables
 */

function Environment(): { [s: string]: string } {
  return Object.keys(process.env).reduce(function (
    o: { [s: string]: string },
    k: string
  ) {
    o['process.env.' + k] = JSON.stringify(process.env[k])
    return o
  }, {})
}

/**
 * Aliases
 */

type Aliases = {
  [s: string]: string
}

// NOTE: these are required server-side
// so the server will require them directly
// and know where they are.
// (e.g. require('elmo/router'))
function ServerAliases(projectRoot: string, elmoRoot: string): Aliases {
  // resolve the project's preact first
  // then look for elmo's preact if
  // we can't find it
  const preact = resolves(
    join(projectRoot, 'node_modules', 'preact'),
    join(elmoRoot, 'node_modules', 'preact'),
  ) || 'preact'

  return {
    react: preact,
    'react-dom': preact,
    'preact': preact,
    'elmo': join(elmoRoot, 'dist', 'src', 'elmo.js'),
    'styled-tsx': join(elmoRoot, 'dist', 'src', 'styled-tsx', 'style.js')
  }
}

// NOTE: these are required client-side
// (e.g. import Router from 'elmo/router')
// these should only be requires that are
// added to the build. the public-facing
// aliases are here to allow us to narrow
// down our elmo typescript loader
function BrowserAliases(_projectRoot: string, elmoRoot: string): Aliases {
  return {
    unfetch: require.resolve('unfetch'),
    'promise-polyfill': require.resolve('promise-polyfill'),
    react: require.resolve('preact'),
    'react-dom': require.resolve('preact'),
    'elmo': join(elmoRoot, 'src', 'elmo.ts'),
    'styled-tsx': join(elmoRoot, 'src', 'styled-tsx', 'style.ts')
  }
}

/**
 * Webpack Config
 */

export default async function Webpack(params: Params) {
  const { path: pkgPath } = await readPkg({ cwd: __dirname })
  const ELMO_ROOT = dirname(pkgPath)
  const ELMO_JS_SRC = join(ELMO_ROOT, 'dist', 'src')
  const ELMO_TS_SRC = join(ELMO_ROOT, 'src')

  const { target, environment, action, routes } = params
  const root = resolve(process.cwd(), params.root || '.')
  const outpath = resolve(root, params.outpath || 'build')

  const serverAliases = ServerAliases(root, ELMO_ROOT)
  const browserAliases = BrowserAliases(root, ELMO_ROOT)

  const config: webpack.Configuration = {
    context: root,
    module: { rules: [] },
    plugins: [],
    entry: {}
  }

  const pages = await globby(join(root, 'pages', '**', '*.{js,jsx,ts,tsx}'))
  const publicPages = pages.filter((page: string) => basename(page)[0] !== '_')

  // not used right now, but perhaps useful in the future
  const common: string[] = []

  // setup the entries
  config.entry = publicPages.reduce(
    (obj: { [s: string]: string[] }, page: string) => {
      // slice off first "/"
      const route = Route(root, page).slice(1)
      const name = join(route, 'index.js')
      let bundle = ([] as string[]).concat(common)

      // only include webpack hot middleware when we're
      // not in production and using the dev server
      // and we're building for the browser
      if (
        action === 'serve' &&
        environment !== 'production' &&
        target === 'browser'
      ) {
        bundle = bundle.concat(
          'webpack-hot-middleware/client?path=/__webpack_hmr&noInfo=true&name=' +
          name
        )
      }

      obj[name] = bundle.concat(page)
      return obj
    },
    {}
  )

  if (target !== 'browser') {
    config.target = 'node'

    // this is to ensure that we use node's actual require(...)
    // which is needed for styled-jsx, but should probably be the
    // default anyway.
    config.externals = [
      function (_parent, request, fn) {
        if (serverAliases[request]) {
          return fn(null, serverAliases[request])
        } else if (request[0] !== '.' && request[0] !== '/') {
          // resolve server-side node modules relative to 
          // the project's root before trying elmo's root.
          // this matters when we run npm link elmo
          let node_module = resolves(
            join(root, 'node_modules', request),
            join(ELMO_ROOT, 'node_modules', request)
          )
          if (!node_module) {
            return fn(new Error('couldn\'t find: ' + request), undefined)
          }

          return fn(null, node_module)
        } else {
          // pass relative paths through (don't externalize)
          return fn(null, undefined)
        }
      }
    ]
  }

  if (target === 'browser') {
    config.node = {
      console: false,
      global: false,
      process: false,
      __filename: true,
      __dirname: true,
      Buffer: false,
      setImmediate: false,
      dns: false,
      fs: false,
      path: true,
      url: false
    }
  }

  // get the document page
  const documentPage =
    pages.find(p => basename(p, extname(p)) === '_document') ||
    join(ELMO_JS_SRC, 'lib', 'document.js')

  // add the document as an entry if it exists
  if (target === 'server') {
    config.entry['document.js'] = common.concat(documentPage)
  }

  // resolves
  config.resolve = {
    extensions: ['.ts', '.tsx', '.js', '.jsx'],
    modules: [join(root, 'node_modules'), join(ELMO_ROOT, 'node_modules')],
    alias: browserAliases
  }

  // where to resolve loaders
  config.resolveLoader = {
    modules: [join(root, 'node_modules'), join(ELMO_ROOT, 'node_modules')]
  }

  /**
   * Output
   */

  config.output = {
    publicPath: '/',
    filename: '[name]',
    path: outpath,
    libraryTarget: 'umd'
  }

  /**
   * Rules
   */

  const rules = (config.module as webpack.NewModule).rules

  // support json files
  rules.push({
    test: /\.json$/,
    use: ['json-loader']
  })

  // compile our project's typescript files
  rules.push({
    test: /\.tsx?$/,
    exclude: /(node_modules|bower_components)/,
    loader: join(ELMO_JS_SRC, 'server', 'loaders', 'typescript.js'),
    query: { environment, target }
  })

  // certain typescript files are allowed to be compiled
  // by the frontend configuration. allow certain
  // modules to be compiled from elmo's ts source
  rules.push({
    test: /\.tsx?$/,
    include: [
      join(ELMO_TS_SRC, 'elmo.ts'),
      join(ELMO_TS_SRC, 'lib'),
      join(ELMO_TS_SRC, 'client'),
      join(ELMO_TS_SRC, 'styled-tsx')
    ],
    loader: join(ELMO_JS_SRC, 'server', 'loaders', 'typescript.js'),
    query: { environment, target }
  })

  // always polyfill to find bugs faster
  // right now just polyfilling fetch and promise
  if (target === 'browser') {
    rules.push({
      test: /\.[tj]sx?$/,
      include: publicPages.map((page: string) => resolve(root, page)),
      use: [join(ELMO_JS_SRC, 'server', 'loaders', 'polyfills.js')]
    })
  }

  // browser-side loaders
  if (target === 'browser') {
    // boot script for client-side, only added to the top-level pages
    rules.push({
      test: /\.[jt]sx?$/,
      include: publicPages.map(page => resolve(root, page)),
      use: {
        loader: join(ELMO_JS_SRC, 'server', 'loaders', 'boot.js'),
        options: {
          context: root
        }
      }
    })
  }

  /**
   * Plugins
   */

  // plugins
  const plugins = config.plugins || []

  // get all the environment variables
  let constants = Environment()

  // attach the environment to the NODE_ENV
  constants['process.env.NODE_ENV'] = JSON.stringify(
    constants['process.env.NODE_ENV'] || environment
  )

  // add the routes into the router
  if (target === 'browser') {
    constants = Object.assign(constants, {
      'window._ELMO_ROUTES_': JSON.stringify(
        routes.length ? routes : await Routes(root)
      )
    })
  }

  // Define environment variables
  plugins.push(new webpack.DefinePlugin(constants))

  // Needed this plugin to detect a `--watch` mode
  // for typescript files
  // if (environment === 'development' && action === 'serve') {
  //   plugins.push(new CheckerPlugin())
  // }

  // Don't write files if there are errors
  plugins.push(new webpack.NoEmitOnErrorsPlugin())

  // use names instead of IDs
  plugins.push(new webpack.NamedModulesPlugin())

  // server-side
  if (target === 'server') {
    plugins.push(
      new StaticSiteGeneratorPlugin(
        {
          elmoRoot: ELMO_ROOT,
          routes,
          environment,
          outpath,
          target,
          root,
          action
        },
        {
          'document.js': documentPage
          // TODO add error & 404 page
        }
      )
    )
  } else if (target === 'browser') {
    plugins.push(
      new webpack.optimize.CommonsChunkPlugin({
        name: 'common',
        filename: 'common.js'
      })
    )
  }

  // copy static assets over when we're building for the browser
  const hasAssets = await exists(join(root, 'static'))
  if (target === 'browser' && hasAssets) {
    plugins.push(
      new CopyWebpackPlugin([
        {
          context: root,
          from: './static',
          to: './static',
          ignore: ['.*']
        }
      ])
    )
  }

  // when serving, show friendly errors
  if (action === 'serve' && target === 'browser') {
    plugins.push(new FriendlyErrors())
  }

  // enable hot module replacement when in dev server in the browser
  if (
    action === 'serve' &&
    environment !== 'production' &&
    target === 'browser'
  ) {
    plugins.push(new webpack.HotModuleReplacementPlugin())
  }

  // minify scripts in production in the browser
  if (environment === 'production' && target === 'browser') {
    plugins.push(
      new webpack.LoaderOptionsPlugin({
        minimize: true,
        debug: false
      })
    )

    plugins.push(
      new webpack.optimize.UglifyJsPlugin({
        beautify: false,
        mangle: {
          screw_ie8: true,
          keep_fnames: true
        },
        compress: {
          warnings: false,
          screw_ie8: true,
          conditionals: true,
          unused: true,
          comparisons: true,
          sequences: true,
          dead_code: true,
          evaluate: true,
          if_return: true,
          join_vars: true,
          negate_iife: false
        },
        sourceMap: false
      })
    )
  }

  config.devtool = environment === 'development' ? 'cheap-source-map' : false

  return config
}

function resolves(...paths: string[]): string | null {
  for (let i = 0; i < paths.length; i++) {
    const path = paths[i]
    try {
      return require.resolve(path)
    } catch (_e) { }
  }
  return null
} 