import Route from '../../lib/route.js'
import { loader } from 'webpack'

export default function (
  this: loader.LoaderContext,
  source: string,
  sourceMap: string
) {
  this.cacheable()

  const route = getRoute(this)
  if (!route) {
    return this.callback(new Error("couldn't find route for page"))
  }

  // This is a bit confusing because we only want to update once, but
  // need to satisfy 3 update conditions: 1st page load, HMR update,
  // and Router.push(...). We only want to update once when we load
  // the page and when there's an HMR update, but not when we async
  // re-route.
  //
  //   -> only do HMR in development mode
  //   if (module.hot) {
  //   -> accept the hot module, this is a noop when we're booting
  //     module.hot.accept()
  //
  //     -> if we're booting, this will be "idle", "apply" occurs
  //     -> when we've made a change, this will allow us to only
  //     -> target HMR updates
  //     if (module.hot.status() === 'apply') {
  //       const Router = require('elmo/router').default
  //       Router.update(Component, route).catch(console.error)
  //       return
  //     }
  //   }
  //
  //   -> This will not be triggered when we're using Router.push(...)
  //   -> there is no corresponding <script>.
  //   if (document.querySelector('script[src="${route}"]')) {
  //     const Router = require('elmo/router').default
  //     Router.update(Component, route).catch(console.error)
  //   }
  this.callback(
    null,
    `
  ${source}
  ;(function (Component, route) {
    Component = Component.default || Component

    if (module.hot) {
      module.hot.accept()

      // this should only occur when we are using
      // HMR, never when we boot or when we use Router.push(...)
      if (module.hot.status() === 'apply') {
        var Router = require('elmo').Router
        Router.update(Component, route).catch(console.error)
        return
      }
    }

    // this should only get rendered on boot, never
    // during HMR or when we use Router.push(...)
    if (document.querySelector('script[src="${route}"]')) {
      var Router = require('elmo').Router
      Router.update(Component, route)
        .then(function () {
          // this will signal to prerender.io compatible
          // clients that the page has rendered successfully
          window.prerenderReady = true
        })
        .catch(console.error)
    }
  })(require('${this.resourcePath}'), ${JSON.stringify(route)})
  `,
    sourceMap
  )
}

function getRoute(loaderContext: loader.LoaderContext): string {
  const root = loaderContext.options.context
  const { resourcePath } = loaderContext
  const dir: string | void = [root].find(d => resourcePath.indexOf(d) === 0)
  if (!dir) return ''
  const route = Route(dir, resourcePath)
  if (route === '/') return '/index.js'
  return route + '/index.js'
}
