import { loader } from 'webpack'
import * as ts from 'typescript'
import Transform from '../../styled-tsx/transform'
import { validate } from '../config/tsconfig'
// import { compile } from 'pylo/server'
// const compiler: any = require('pylo/server').compile.default
// console.log(compiler.compile.default)
// compiler as any
import { join } from 'path'
// const SRC_ROOT = join(__dirname, '..', '..')

export default function(
  this: loader.LoaderContext,
  source: string,
  _sourceMap: string
) {
  this.cacheable()

  const { resourcePath, options: { context: root } } = this
  const { environment, target } = this.query

  // this is the project's tsconfig, not elmo/s
  const TSConfig = possRequire(join(root, 'tsconfig.json'))
  if (TSConfig instanceof Error) {
    const err = new Error(
      'invalid or missing tsconfig.json at: ' + join(root, 'tsconfig.json')
    )
    this.callback(err)
    return
  }

  const errors = validate(TSConfig)
  if (errors.length) {
    const err = new TypeError(
      '\n' + errors.map(e => e.message).join('\n') + '\n'
    )
    this.callback(err)
    return
  }

  const hasRuntime = environment !== 'production' || target !== 'browser'

  // console.log('transpiling %s %s', root, resourcePath)
  const result = ts.transpileModule(source, {
    fileName: resourcePath,
    reportDiagnostics: true,
    transformers: {
      before: [
        visitor(
          Transform({
            runtime: hasRuntime
          })
        )
      ]
    },
    compilerOptions: TSConfig.compilerOptions
  })

  const err = error(result.diagnostics)
  if (err) {
    this.callback(err)
    return
  }

  this.callback(null, result.outputText, result.sourceMapText)
}

// create an error if it exists
function error(diagnostics: ts.Diagnostic[] = []): TypeError | void {
  const errors = diagnostics.map(diagnostic => {
    if (!diagnostic.file || !diagnostic.start) return
    const { line, character } = diagnostic.file.getLineAndCharacterOfPosition(
      diagnostic.start
    )
    const reason = ts.flattenDiagnosticMessageText(diagnostic.messageText, '\n')
    return `${diagnostic.file.fileName} (${line + 1},${character +
      1}): ${reason}`
  })

  if (!errors.length) return
  return new TypeError('\n' + errors.join('\n') + '\n')
}

type Visitors<T> = {
  [type: number]:
    | Visitor<T>
    | {
        enter?: Visitor<T>
        exit?: Visitor<T>
      }
}

type Visitor<T> = (node: T, context?: ts.TransformationContext) => T | void

function visitor(transform: Visitors<ts.Node>) {
  return (context: ts.TransformationContext) => {
    return (sourceFile: ts.SourceFile): ts.SourceFile => {
      return visit(sourceFile)
      function visit<T extends ts.Node>(node: T): T {
        const t = transform[node.kind]
        if (!t) {
          return ts.visitEachChild(node, visit, context)
        }

        // enter a node
        if (typeof t === 'object' && t.enter) {
          const tr = t.enter(node, context)
          if (tr) node = tr as T
        }

        // if it's a function, treat it
        // the same way as t.enter
        if (typeof t === 'function') {
          const tr = t(node, context)
          if (tr) node = tr as T
        }

        // visit each child
        let exitNode = ts.visitEachChild(node, visit, context)

        // if we have an exit call function
        if (typeof t === 'object' && t.exit) {
          const tr = t.exit(exitNode, context)
          if (tr) exitNode = tr as T
        }

        return exitNode
      }
    }
  }
}

function possRequire(req: string): { [key: string]: any } | Error {
  try {
    return require(req)
  } catch (e) {
    return e
  }
}
