import { Route } from '../../elmo'
import { h, VNode } from 'preact'
import * as Stringify from 'preact-render-to-string'
import { Plugin, Compiler } from 'webpack'
import * as Evaluate from 'eval'
import { join, extname, basename, dirname, resolve } from 'path'
import flush from '../../styled-tsx/server'

export type Entries = {
  [s: string]: {
    source: () => string
    size: () => number
  }
}

export type EntryPaths = {
  [entry: string]: string
}

export type StaticParams = {
  environment: string
  elmoRoot: string
  outpath: string
  target: string
  action: string
  root: string
  routes: Route[]
  port?: number
}

/**
 * Module Exports
 */

export default class Static implements Plugin {
  private readonly options: StaticParams
  private readonly entrypaths: EntryPaths

  constructor(options: StaticParams, entrypaths: EntryPaths) {
    this.options = options
    this.entrypaths = entrypaths
  }

  async build(entries: Entries) {
    const documentPath = this.entrypaths['document.js']
    const { elmoRoot, root } = this.options

    // evaluate our document.js file
    const document: preact.Component<
      Object,
      null
      > = evaluate(documentPath, entries['document.js'].source(), { require })

    if (typeof document !== 'function') {
      throw new Error(
        '_document should be a Component. Please make sure you exported it'
      )
    }

    // filter out the pages
    const pages = Object.keys(entries)
      .filter(entry => /\.[tj]sx?/.test(extname(entry)))
      .filter(entry => /index\.[tj]sx?/.test(entry))

    // asynchronously evaluate and get build props
    const pageData = await Promise.all(
      pages.map(async filepath => {
        // evaluate the page server-side
        const source = entries[filepath].source()
        // evaluate from the perspective of the entry file
        // NOTE: require is important here for styled-jsx flush() to work
        // correctly, we need to use a shared require statement to share
        // the module cache
        const Component = evaluate(resolve(root, filepath), source, { require })

        // get the build properties, if there are any
        const props = await getBuildProps(Component)

        return { filepath, props, source, Component }
      })
    )

    // render HTML synchronously to avoid any race conditions
    // with rewind
    await Promise.all(pageData.map(async ({ filepath, props, Component }) => {
      // render the server-side HTML
      const html = await renderHTML(document, {
        elmoRoot,
        filepath,
        Component,
        props
      })

      addFile(entries, filepath, html)
    }))

    // we go ahead and delete all the JS files,
    // because this is a server-side build and
    // we only care about the client-side build
    // results
    for (var entryPath in entries) {
      if (extname(entryPath) !== '.js') continue
      delete entries[entryPath]
    }
  }

  apply(compiler: Compiler) {
    const self = this

    compiler.plugin('this-compilation', function (compiler: Compiler) {
      compiler.plugin('optimize-assets', function (
        entries,
        done: (err?: Error) => void
      ) {
        self
          .build(entries)
          .then(() => done())
          .catch(e => {
            done(e)
          })
      })
    })
  }
}

/**
 * Get the initial properties of the static build
 * // TODO: better inputs
 */
function getBuildProps(Component: any): Promise<Object> {
  if (Component.getBuildProps) {
    return Component.getBuildProps({ build: true, path: '', query: {} })
  } else {
    return Promise.resolve({})
  }
}

/**
 * Render the HTML
 */

// TODO: not exactly sure how to type the Document when there's an evaluate()
async function renderHTML(
  Document: any,
  params: {
    elmoRoot: string
    filepath: string
    Component: any
    props: any
  }
) {
  const Component = params.Component
  const page = Stringify(<Component {...params.props} />)
  const { head, style } = rewindAll(params.elmoRoot)

  const docProps = await getBuildProps(Document)

  // create a virtual DOM tree
  const vhtml = (
    <Document
      page={page}
      head={head}
      style={style}
      props={docProps || {}}
      filepath={params.filepath}
    />
  )

  const html = '<!DOCTYPE html>' + Stringify(vhtml)

  // ignore any <style scoped> or <Head /> in the document
  rewindAll(params.elmoRoot)

  return html
}

/**
 * Add the entry file
 */

function addFile(entries: Entries, entry: string, html: string) {
  const name = basename(dirname(entry))
  entries[name + '/index.html'] = {
    source: () => html,
    size: () => html.length
  }
  return null
}

/**
 * Evaluate the component
 */

function evaluate(
  entry: string,
  source: string,
  globals?: { [key: string]: any }
) {
  var render
  render = Evaluate(
    source,
    /* filename: */ entry,
    /* scope: */ globals,
    /* includeGlobals: */ true
  )
  if (render.hasOwnProperty('default')) {
    render = render['default']
  }
  return render
}

// Okay so the reason for this is that when we call evaluate
// it may use a non-cached version of <Head /> or <Style/>
// depending on which module we require. These need to be
// pull from the cache to maintain their flush state.

// This function normalizes over all possible places
// you can get these modules from, so no matter where
// we require it we'll get the state.
function rewindAll(
  elmoRoot: string
): { head: VNode<any>[]; style: VNode<any>[] } {
  const dist = join(elmoRoot, 'dist', 'src')
  const src = join(elmoRoot, 'src')
  const cache = require.cache

  const distHead = join(dist, 'lib', 'head.js')
  const srcHead = join(src, 'lib', 'head.ts')
  const distStyle = join(dist, 'styled-tsx', 'style.js')
  const srcStyle = join(src, 'styled-tsx', 'style.ts')
  let fn = null

  let head: VNode<any>[] = []
  let style: VNode<any>[] = []

  for (let req in cache) {
    if (!cache[req]) continue
    switch (req) {
      case distHead:
        fn = cache[req].exports.default
        head = head.concat(fn.rewind())
        break
      case srcHead:
        fn = cache[req].exports.default
        head = head.concat(fn.rewind())
        break
      case distStyle:
        fn = cache[req].exports.default
        style = style.concat(flush(fn.flush()))
        break
      case srcStyle:
        fn = cache[req].exports.default
        style = style.concat(flush(fn.flush()))
        break
    }
  }

  return { head, style }
}
